/***************************************************************************
                          taxi.cpp  -  description
                             -------------------
    begin                : Fri May 18 2001
    copyright            : (C) 2001, 2002, 2004 by Thomas Friedrichsmeier
    email                : Thomas.Friedrichsmeier@ruhr-uni-bochum.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include "defines.h"

#include <stdlib.h>

#include <qdom.h>

#include "taxi.h"

#include "taxipilot.h"
#include "cdp.h"
#include "teleporter.h"
#include "platform.h"
#include "movingobject.h"
#include "soundwrapper.h"
#include "animation.h"

Taxi::Taxi(const QString & file, int base_x, int base_y, bool for_credits_only):AnimatedSprite()
{

	idle_timer = anim_timer = flaps_anim_timer = advance_timer =-1;
	flying_lag_x = flying_lag_y = 0;
	flying_state = prev_flying_state = 0;
	fuel_low = fuel_out = false;
	flaps_state = 0;
	flaps_avail = flaps_prev_state = false;
	landed = false;

	flaps = new AnimatedSprite();
	flaps->setZ(TAXI_FLAPS_Z);
	flaps->setAnimated (false);

	// prepare the flying-animations
	flying_lr = new AnimatedSprite ();
	flying_lr->setAnimated (false);
	flying_ud = new AnimatedSprite ();
	flying_ud->setAnimated (false);
	flying_lr->setZ(TAXI_FLYING_Z);
	flying_ud->setZ(TAXI_FLYING_Z);

	read_taxi(file);

	// intialize flying-animation with a dummy (just to be sure)
	if (cdp->for_real()) {
		flying_lr->setAnimation (flying_left);
		flying_ud->setAnimation (flying_up);
        }

/* calculate offset x and y */
	setAnimation (still);
	off_x = width() / 2;
	if (flaps_avail && cdp->for_real()) {
		twice_off_y = flaps->boundingRect().height();
	} else {
		twice_off_y = height();
	}
	off_y = twice_off_y / 2;

//	paused = false;
	setZ(TAXI_Z);
	dynamic_teleport = false;

// appear!
	if (for_credits_only) {
		go_idle ();
	} else {
		teleport_to(base_x, base_y, 0, 0, true);
		show();					// here we go! ( Note that teleport_to only triggers teleporting! )
		advance_timer = tm->startTimer (this, cdp->framePeriod ());
	}

// the taxi is the major moving object. Therefore we adjust the canvas to it.
	int x = width ();   				// max (width (), height ())
	if (height () > x) {
		x = height ();
	}
	TPBase::canvas->retune (16 + ((int) (x / 16)) * 16);

	config_changed ();
}

Taxi::~Taxi()
{

	tm->killTimers (this);
	hide ();
	flaps->hide ();
	flying_lr->hide ();
	flying_ud->hide ();

	for (unsigned int i = 0; i < num_idle_animations; i++) {
		delete idle[i];
	}

	delete teleporting;
	delete still;

	if (cdp->for_real () || cdp->thorough_check ()) {
		delete flying_up;
		delete flying_down;
		delete flying_left;
		delete flying_right;
		if (flaps_avail) {
			delete flaps_anim;
		}
	}

	delete[]idle;
	delete flying_lr;
	delete flying_ud;
	delete flaps;

}

/** Signals we are going up. The reason this is not simply done via setVelocity is
that taxi is eventually to do animation stuff. The float acceleration is passed as
a parameter instead of using some constant amount, so we may eventually
abstract from keyboard input, and allow (analog) joystick input */
void Taxi::up(float acc)
{
	if (ready) {
		if (fuel > 0) {

			if (!(prev_flying_state & Up)) {
				flying_ud->setAnimation (flying_up);
				flying_ud->startSound (1);
				flying_lag_y = 0.0;
				flying_state -= (flying_state & Down);
			}
			flying_state |= Up;

			flying_ud->tick ();
			if (flying_lag_y < MAX_FLYING_LAG) flying_lag_y += flying_lag_step;
			flying_ud->show();

			if (landed) {
				landed = false;
				platform_p->detach_taxi ();
				setXVelocity ((platform_p->current_plat_vx () * (cdp->framePeriod () / OBJECTS_AP)));
				setYVelocity ((platform_p->current_plat_vy () * (cdp->framePeriod () / OBJECTS_AP)));
				wrap_moveBy (0, -2);		// To avoid getting "caught" again right away
			}

			setYVelocity(yVelocity() + acc);

			fuel -= fuel_con;
		}
	}
}

/** Signals we are going down. See up () for details. */
void Taxi::down(float acc)
{
	if (ready && !landed) {
		if (fuel > 0) {

			if (!(prev_flying_state & Down)) {
				flying_ud->setAnimation (flying_down);
				flying_ud->startSound (1);
				flying_lag_y = 0.0;
				flying_state -= (flying_state & Up);
			}
			flying_state |= Down;

			flying_ud->tick ();
			if (flying_lag_y > (-MAX_FLYING_LAG)) flying_lag_y -= flying_lag_step;
			flying_ud->show();

			setYVelocity(yVelocity() + acc);

			fuel -= fuel_con;
		}
	}
}

/** See above */
void Taxi::left(float acc)
{
	if (ready && !(flaps_block && flaps_state) && !landed) {
		if (fuel > 0) {

			if (!(prev_flying_state & Left)) {
				flying_lr->setAnimation (flying_left);
				flying_lr->startSound (1);
				flying_lag_x = 0.0;
				flying_state -= (flying_state & Right);
			}
			flying_state |= Left;

			flying_lr->tick ();
			if (flying_lag_x < MAX_FLYING_LAG) flying_lag_x += flying_lag_step;
			flying_lr->show();

			if (flaps_avail) {
				flaps->setOrientation (AnimLeft);
			}
			setOrientation (AnimLeft);
			flying_ud->setOrientation (AnimLeft);

			setXVelocity(xVelocity() + acc);

			fuel -= fuel_con;
		}
	}
}

/** See above */
void Taxi::right(float acc)
{
	if (ready && !(flaps_block && flaps_state) && !landed) {
		if (fuel > 0) {

			if (!(prev_flying_state & Right)) {
				flying_lr->setAnimation (flying_right);
				flying_lr->startSound (1);
				flying_lag_x = 0.0;
				flying_state -= (flying_state & Left);
			}
			flying_state |= Right;

			flying_lr->tick ();
			if (flying_lag_x > (-MAX_FLYING_LAG)) flying_lag_x -= flying_lag_step;
			flying_lr->show();

			if (flaps_avail) {
				flaps->setOrientation (AnimRight);
			}
			setOrientation (AnimRight);
			flying_ud->setOrientation (AnimRight);

			setXVelocity(xVelocity() + acc);

			fuel -= fuel_con;
		}
	}
}

/** This function reads a taxi-description ( specs for the animations ) from a file.
read_taxi is only called by the constructor. */
void Taxi::read_taxi(const QString & file)
{

	QDomElement e;				// just some element
	QDomNodeList list;			// just a list of nodes

	QDomElement xdocElement = cdp->open_config_file(file);

	author = cdp->get_string_attribute("author", xdocElement, i18n("Anonymous"), Info);
	title = cdp->get_string_attribute("title", xdocElement, i18n("Unnamed"), Info);

	e = cdp->get_element("teleporting", xdocElement, Critical);	// find the teleporting-animation
	teleporting = new Animation (&e, ReadSound, 0, "_left", "_right");

	e = cdp->get_element("still", xdocElement, Critical);	// find the standing-still-image/animation
	still = new Animation (&e, ReadSound | DefaultAnimContinuous | ReadAnimOscillating, ReadContinuous | ReadInterruptable, "_left", "_right");

	e = cdp->get_element("idle", xdocElement, Critical);	// find the idle_animations-section
	list = cdp->get_node_list("data", e, 1, -1, 1, Critical);	// get all idle-animations

	num_idle_animations = 0;
	idle = new Animation * [list.length()];

	while (num_idle_animations < list.length ()) {
		e = list.item (num_idle_animations).toElement();

		idle[num_idle_animations] = new Animation (&e, ReadSound | ReadAnimRepetitions | ReadAnimOscillating, DefaultSingleSound | ReadRepetitious, "_left", "_right");

		num_idle_animations++;
	}

	if (!(cdp->for_real () || cdp->thorough_check ())) {
		return;		// We don't need all the other stuff for the start-screen
	}

	// read global sounds
	e = cdp->get_element("crash", xdocElement, Info);
	crash_sound.create (e, DefaultSingleSound, Info);

	e = cdp->get_element("land", xdocElement, Info);
	land_sound.create (e, DefaultSingleSound, Info);

	e = cdp->get_element("flying_left", xdocElement, Critical);	// find the flying_left/right-animation
	flying_left = new Animation (&e, ReadSound | DefaultAnimContinuous | ReadAnimOscillating, DefaultSingleSound | DefaultContinuous | ReadInterruptable | DefaultInterruptable);
	e = cdp->get_element("flying_right", xdocElement, Critical);
	flying_right = new Animation (&e, ReadSound | DefaultAnimContinuous | ReadAnimOscillating, DefaultSingleSound | DefaultContinuous | ReadInterruptable | DefaultInterruptable);

	e = cdp->get_element("flying_up", xdocElement, Critical);	// find the flying_up/down-animation
	flying_up = new Animation (&e, ReadSound | DefaultAnimContinuous | ReadAnimOscillating, DefaultSingleSound | DefaultContinuous | ReadInterruptable | DefaultInterruptable, "_left", "_right");
	e = cdp->get_element("flying_down", xdocElement, Critical);
	flying_down = new Animation (&e, ReadSound | DefaultAnimContinuous | ReadAnimOscillating, DefaultSingleSound | DefaultContinuous | ReadInterruptable | DefaultInterruptable, "_left", "_right");

	e = cdp->get_element("landing_flaps", xdocElement, Warn);

	setAnimation (still);		// needed for width () below

	if (e.hasAttribute("offset_x1_left") || e.hasAttribute("offset_x1_right")) {
		landing_flaps_off_x1[AnimLeft] = cdp->get_int_attribute("offset_x1_left", e, -20, width() + 20, 0, Warn);
		landing_flaps_off_x1[AnimRight] = cdp->get_int_attribute("offset_x1_right", e, -20, width() + 20, 0, Warn);
	} else {
		landing_flaps_off_x1[AnimLeft] = landing_flaps_off_x1[AnimRight] =
			cdp->get_int_attribute("offset_x1", e, -20, width() + 20, 0, Warn);
	}
	if (e.hasAttribute("offset_x2_left") || e.hasAttribute("offset_x2_right")) {
		landing_flaps_off_x2[AnimLeft] = cdp->get_int_attribute("offset_x2_left", e, -20, width() + 20, width(), Warn);
		landing_flaps_off_x2[AnimRight] = cdp->get_int_attribute("offset_x2_right", e, -20, width() + 20, width(), Warn);
	} else {
		landing_flaps_off_x2[AnimLeft] = landing_flaps_off_x2[AnimRight] =
			cdp->get_int_attribute("offset_x2", e, -20, width() + 20, width(), Warn);
	}

	if ((flaps_avail = cdp->get_bool_attribute("moveable", e, false, Info))) {
		flaps_anim = new Animation (&e, ReadSound, 0, "_left", "_right");
		flaps->setAnimation (flaps_anim);
		flaps_block = cdp->get_bool_attribute("blocking", e, false, Info);
	} else {
		flaps_block = false;
		flaps_state = 2;		// if flaps are not moveable, they are always out.
	}

}

/** Stops the currently running flying animation, if any (called from Taxipilot::do_chores ()) */
void Taxi::stop_flying_animation()
{
	int sounds_to_stop = prev_flying_state & (~flying_state);

	if (sounds_to_stop) {
		if (sounds_to_stop & (Up | Down)) {
			if (!(flying_state & (Up | Down))) {
				flying_ud->hide();
				flying_ud->stopSound ();
			}
		}
		if (sounds_to_stop & (Left | Right)) {
			if (!(flying_state & (Left | Right))) {
				flying_lr->hide();
				flying_lr->stopSound ();
			}
		}
	}

	prev_flying_state = flying_state;
	flying_state = 0;
}

  /** Takes care of teleportation (appearing, disappearing animations), setting new position
and speed. If appear_only is given, does not do disappearing animation. */
void Taxi::teleport_to(int x, int y, float speedfactor_x, float speedfactor_y, bool appear_only, Teleporter *dynamic)
{

	if (landed) platform_p->detach_taxi ();
	ready = landed = false;

	if (!flaps_state) {			// make sure, flaps show up correctly after teleporting
		flaps_prev_state = false;
	} else if (flaps_state == 2) {
		flaps_prev_state = true;
	} else {
		if (flaps->getDirection () == AnimReverse) {
			flaps_prev_state = false;
		} else {
			flaps_prev_state = true;
		}
	}

	if (dynamic) {
		dynamic_teleporter = dynamic;
		dynamic_teleport = true;
	} else {
		dynamic_teleport = false;
		new_x = x - off_x;
		new_y = y - off_y;
	}

	new_speed_x = xVelocity() * speedfactor_x;
	new_speed_y = yVelocity() * speedfactor_y;
	setVelocity(0, 0);

	if (idle_timer != -1) {
		tm->killTimer(idle_timer);
		idle_timer = -1;
	}

	if (appear_only) {
		reappear ();
	} else {
		do_animation (teleporting, &Taxi::reappear);
		do_animation (flaps_anim, &Taxi::flaps_in, false, true, true);
	}

}

/** Returns the current fuel level in percent of the taxi (Used by Taxipilot to set the
FuelDial) */
int Taxi::fuel_state()
{
	int return_val;
	return_val = (int) (((float) fuel / max_fuel) * 100);
	if (return_val < 10) {
		fuel_low = true;
		if (fuel <= 0) {
			fuel_out = true;
		}
	} else {
		fuel_low = fuel_out = false;
	}
	return (return_val);
}

/** Increases the taxi's fuel supply by step (max up to max_fuel). Returns actual
increase in fuel (so that's 0 if taxi is already full). */
float Taxi::fuel_up(int step)
{
	float prev_fuel;				// a dummy storing the previous fuel state

	if (step <= 0)
		return (0);				// no negative fueling, please

	prev_fuel = fuel;
	fuel += (step * fuel_con);
	if (fuel > max_fuel)
		fuel = max_fuel;
	return (fuel - prev_fuel);
}

/** Initializes the taxi to have fuel fuel and a fuel-capacity max_fuel */
void Taxi::initialize (float fuel, float max_fuel) {
	Taxi::fuel = fuel;
	Taxi::max_fuel = max_fuel;
}


/** Toggles the taxis landing-flaps */
void Taxi::toggle_flaps()
{

	if (flaps_avail && ready) {	// if not, we simply ignore this
		if (flaps_state == 1) {	// flaps are already moving some way
			if (flaps->getDirection () == AnimReverse) {		// so we need to reverse this
				flaps->setDirection (AnimForward);
			} else {
				flaps->setDirection (AnimReverse);
			}
		} else {
			if (flaps_state == 0) {	// flaps are in, going out
				do_animation (flaps_anim, &Taxi::flaps_out, true, true);
			} else {			// flaps are out, going in
				if (landed) {
					platform_p->detach_taxi ();
					landed = false;
				}
				do_animation (flaps_anim, &Taxi::flaps_in, false, true);
			}
		}
	}

}

/** Repositions the taxi. This also takes care of repositioning the landing flaps. */
void Taxi::reposition(int x, int y, float xv, float yv, bool align_bottom)
{

	int temp_off_y;

	if (align_bottom) {
		temp_off_y = twice_off_y;
	} else {
		temp_off_y = off_y;
	}

	if (x < 0) {				// position parameter = -1 means don't change position.
		if (y >= 0) {
			moveBy(0, (y - temp_off_y) - Taxi::y());
			flaps->moveBy(0, (y - temp_off_y) - flaps->y());
		}
	} else {
		if (y < 0) {
			moveBy(x - Taxi::x(), 0);
			flaps->moveBy(x - flaps->x(), 0);
		} else {
			move(x, (y - temp_off_y));
			flaps->move(x, (y - temp_off_y));
		}
	}

	setVelocity(xv, yv);
}

/** Tells the taxi it crashed. Does stuff like setting of an appropriate animation,
sound and other things to do. hard specifies, whether the taxi crashed into something hard (as opposed to
into the passenger) and whether the crash_sound will be played */
void Taxi::crash(bool hard)
{
	setVelocity(0, 0);
	if (flaps_state) {
		do_animation (flaps_anim, &Taxi::flaps_in, false, true, true);
	}
	if (landed) {
		landed = false;
		platform_p->detach_taxi ();
	}

	flying_lr->stopSound ();
	flying_ud->stopSound ();

	if (hard) {
		sw->start_sound (crash_sound.id[0]);
	}
}

/** Returns the (absolute) x-position the left landing-flaps is to be found at */
int Taxi::landing_flaps_x1()
{
	return ((int) x() + landing_flaps_off_x1[getOrientation ()]);
}

/** Returns the (absolute) x-position the right landing-flap is to be found at */
int Taxi::landing_flaps_x2()
{
	return ((int) x() + landing_flaps_off_x2[getOrientation ()]);
}

/** Returns the (absolute) y-position the landing-flaps are to be found at */
int Taxi::landing_flaps_y(){
	if (!flaps_avail) return (bottomEdge());
	return (flaps->boundingRect().bottom());
}

/** A wrapper for moveBy, taking care of the flaps if appropriate */
void Taxi::wrap_moveBy (float delta_x, float delta_y){
	moveBy (delta_x, delta_y);
	flaps->moveBy (delta_x, delta_y);
}

/** Adjusts the taxi to a fractional position, so it doesn't wiggle on a moving platform */
void Taxi::adjust (float fraction_x, float fraction_y) {
	move (((int) x()) + fraction_x, ((int) y()) + fraction_y);
	flaps->move (((int) x()) + fraction_x, ((int) y()) + fraction_y);
}

/** Sets of the animation anim for this taxi or it's flaps, and triggers
a function-call when completed. Reverse means: play in reverse order; flaps means, the animation is for the flaps */
void Taxi::do_animation (Animation * anim, void (Taxi::*call_back) (), bool reverse, bool flaps, bool no_sound) {

	AnimDirections anim_dir = AnimForward;
	if (reverse) {
		anim_dir = AnimReverse;
	}

	if (flaps) {
		if (flaps_avail) {				// no need to bother otherwise
			Taxi::flaps->setAnimation (anim, anim_dir);
			flaps_call_back = call_back;
			if (flaps_anim_timer != -1) {
				tm->killTimer (flaps_anim_timer);
			}
			flaps_anim_timer = tm->startTimer (this, anim->framePeriod ());
			flaps_state = 1;
			Taxi::flaps->show ();
			if (!no_sound) {
				Taxi::flaps->startSound (1);
			}
		}
	} else {
		setAnimation (anim, anim_dir);
		taxi_call_back = call_back;
		if (anim_timer != -1) {
			tm->killTimer (anim_timer);
		}
		anim_timer = tm->startTimer (this, anim->framePeriod ());
		if (!no_sound) {
			startSound (1);
		}
	}

}

/** Stuff to do when the flaps just finished going out */
void Taxi::flaps_out(){
	flaps_state = 2;
}

/** Stuff to do when flaps have just finished going in */
void Taxi::flaps_in (){
	flaps_state = 0;
	flaps->hide();
}

/** Taxi is to reappear (at global new_x, new_y) */
void Taxi::reappear (){

	if (dynamic_teleport) {
		new_x = dynamic_teleporter->destX() - off_x;
		new_y = dynamic_teleporter->destY() - off_y;
	}

	move(new_x, new_y);
	flaps->move(new_x, new_y);

	do_animation (teleporting, &Taxi::done_appearing, true);

	if (flaps_avail && flaps_prev_state) {
		do_animation (flaps_anim, &Taxi::flaps_out, true, true, true);
	}

}

/** Starts idle_timer (not idle_animation) */
void Taxi::go_idle (){
	if (idle_timer != -1) {
		tm->killTimer (idle_timer);		//just in case
	}

	idle_timer = tm->startTimer (this, (rand() % 3000) + 2000);

	do_animation (still, &Taxi::do_nothing);         // this is a continuous animation. no call_back needed
}

/** Does nothing */
void Taxi::do_nothing(){
}

/** Stuff to do when done appearing */
void Taxi::done_appearing (){

	ready = true;	// we're ready to go
	setVelocity(new_speed_x, new_speed_y);
	go_idle ();

}

/** The config has changed. Recalculate. */
void Taxi::config_changed(){
	fuel_con = 20 / (float) cdp->cps ();
	flying_lag_step = MAX_FLYING_LAG / cdp->cps ();
	if (advance_timer != -1) {
		tm->killTimer (advance_timer);
		advance_timer = tm->startTimer (this, cdp->framePeriod ());
	}
}

/** Mostly everyting will be done inside: timing for all animations */
void Taxi::timerTick(int id)
{
	if (id == advance_timer) {	// advance taxi and associated sprites
		moveBy (xVelocity (), yVelocity ());
		flaps->moveBy (xVelocity (), yVelocity ());
		flying_lr->move (x () + flying_lag_x, y ());
		flying_ud->move (x (), y () + flying_lag_y);
	} else if (id == idle_timer) {	// this only sets off the actual animation
		tm->killTimer (idle_timer);
		idle_timer = -1;
		do_animation (idle[(rand() % num_idle_animations)], &Taxi::go_idle);
	} else if (id == anim_timer) {
		if (tick ()) {
			tm->killTimer (anim_timer);
			anim_timer = -1;
			(this->*taxi_call_back) ();
		}
	} else if (id == flaps_anim_timer) {
		if (flaps->tick ()) {
			tm->killTimer (flaps_anim_timer);
			flaps_anim_timer = -1;
			(this->*flaps_call_back) ();
		}
	}
}

/** Tell the taxi it has landed on the specified paltform */
void Taxi::land (Platform *on) {
	sw->start_sound (land_sound.id[0]);
	landed = true;
	platform_p = on;
	platform_p->attach_taxi (this);
}
